from itertools import product

import astropy.units as u
from sunpy.net import attrs as a
from sunpy.net.dataretriever.client import GenericClient, QueryResponse
from sunpy.time import TimeRange
from sunpy.util.scraper import Scraper

from radiospectra.net import attrs as ra

__all__ = ['SWAVESClient']


RECEIVER_FREQUENCIES = {
    'lfr': a.Wavelength(10*u.kHz, 160*u.kHz),
    'hfr': a.Wavelength(0.125*u.MHz, 16*u.MHz),
}


class SWAVESClient(GenericClient):
    """
    Provides access to STEREO `S-WAVES <https://swaves.gsfc.nasa.gov/swaves_instr.html>`__ radio
    data `archive <https://solar-radio.gsfc.nasa.gov/data/stereo/>`__

    Examples
    --------
    >>> import radiospectra.net
    >>> from sunpy.net import Fido, attrs as a
    >>> results = Fido.search(a.Time("2010/10/01", "2010/10/02"),
    ...                       a.Instrument('SWAVES'))  # doctest: +REMOTE_DATA
    >>> print(results[1])  # doctest: +REMOTE_DATA
           Start Time               End Time        ... Provider  Wavelength [2]
                                                    ...                kHz
    ----------------------- ----------------------- ... -------- ----------------
    2010-10-01 00:00:00.000 2010-10-01 23:59:59.999 ...     NASA    10.0 .. 160.0
    2010-10-02 00:00:00.000 2010-10-02 23:59:59.999 ...     NASA    10.0 .. 160.0
    2010-10-01 00:00:00.000 2010-10-01 23:59:59.999 ...     NASA 125.0 .. 16000.0
    2010-10-02 00:00:00.000 2010-10-02 23:59:59.999 ...     NASA 125.0 .. 16000.0
    2010-10-01 00:00:00.000 2010-10-01 23:59:59.999 ...     NASA    10.0 .. 160.0
    2010-10-02 00:00:00.000 2010-10-02 23:59:59.999 ...     NASA    10.0 .. 160.0
    2010-10-01 00:00:00.000 2010-10-01 23:59:59.999 ...     NASA 125.0 .. 16000.0
    2010-10-02 00:00:00.000 2010-10-02 23:59:59.999 ...     NASA 125.0 .. 16000.0

    """

    baseurl = (r'https://solar-radio.gsfc.nasa.gov/data/stereo/summary/{year}/'
               r'swaves_average_(\d){{8}}_{Spacecraft}_{Wavelength}.dat')

    pattern = r'{}/{year:4d}/' \
              r'swaves_average_{year:4d}{month:2d}{day:2d}_{Spacecraft}_{Wavelength}.dat'

    @classmethod
    def _check_wavelengths(cls, wavelength):
        """
        Check for overlap between given wavelength and receiver frequency coverage defined in
        `RECEIVER_FREQUENCIES`.

        Parameters
        ----------
        wavelength : `sunpy.net.attrs.Wavelength`
            Input wavelength range to check

        Returns
        -------
        `list`
            List of receivers names or empty list if no overlap
        """
        # Input wavelength range is completely contained in one receiver range
        receivers = [k for k, v in RECEIVER_FREQUENCIES.items() if wavelength in v]
        # If not defined need to continue
        if not receivers:
            # Overlaps but not contained in, either max in lfr or min hfr
            if wavelength.min in RECEIVER_FREQUENCIES['hfr'] or \
                    wavelength.max in RECEIVER_FREQUENCIES['hfr']:
                receivers.append('hfr')
            if wavelength.min in RECEIVER_FREQUENCIES['lfr'] or \
                    wavelength.max in RECEIVER_FREQUENCIES['lfr']:
                receivers.append('lfr')
            # min in lfr and max in hfr
            # min and max of combined lft and hfr contained in give wavelength range
            if a.Wavelength(RECEIVER_FREQUENCIES['lfr'].min,
                            RECEIVER_FREQUENCIES['hfr'].max) in wavelength:
                receivers = ['lfr', 'hfr']
            # If we get here the is no overlap so set to empty list
        return receivers

    def search(self, *args, **kwargs):
        """
         Query this client for a list of results.

         Parameters
         ----------
         *args: `tuple`
             `sunpy.net.attrs` objects representing the query.
         **kwargs: `dict`
              Any extra keywords to refine the search.

         Returns
         -------
         A `QueryResponse` instance containing the query result.
         """
        matchdict = self._get_match_dict(*args, **kwargs)
        req_wave = matchdict.get('Wavelength', None)
        receivers = RECEIVER_FREQUENCIES.keys()
        if req_wave is not None:
            receivers = self._check_wavelengths(req_wave)

        spacecraft = matchdict.get('spacecraft', ['a', 'b'])

        metalist = []
        start_year = matchdict['Start Time'].datetime.year
        end_year = matchdict['End Time'].datetime.year
        tr = TimeRange(matchdict['Start Time'], matchdict['End Time'])
        for spc, receiver in product(spacecraft, receivers):
            for year in range(start_year, end_year+1):
                urlpattern = self.baseurl.format(Wavelength=receiver, Spacecraft=spc, year=year)
                scraper = Scraper(urlpattern, regex=True)
                filesmeta = scraper._extract_files_meta(tr, extractor=self.pattern)
                for i in filesmeta:
                    rowdict = self.post_search_hook(i, matchdict)
                    metalist.append(rowdict)

        return QueryResponse(metalist, client=self)

    def post_search_hook(self, exdict, matchdict):
        """
        This method converts 'rfs_hfr' and 'rfs_lfr' in the url's metadata
        to the frequency ranges of for low and high frequency receivers.
        """
        rowdict = super().post_search_hook(exdict, matchdict)
        if rowdict['Wavelength'] == 'hfr':
            fr = RECEIVER_FREQUENCIES['hfr']
            rowdict['Wavelength'] = u.Quantity([float(fr.min.value), float(fr.max.value)],
                                               unit=fr.unit)
        elif rowdict['Wavelength'] == 'lfr':
            fr = RECEIVER_FREQUENCIES['lfr']
            rowdict['Wavelength'] = u.Quantity([float(fr.min.value), float(fr.max.value)],
                                               unit=fr.unit)
        return rowdict

    @classmethod
    def register_values(cls):
        adict = {a.Instrument: [('SWAVES', 'STEREO WAVES - SWAVES')],
                 a.Source: [('STEREO', 'Solar Terrestrial Relations Observatory')],
                 ra.Spacecraft: [('A', 'Ahead'), ('B', 'Behind')],
                 a.Provider: [('NASA', 'NASA Goddard Space Flight Center')],
                 a.Wavelength: [('*')]}
        return adict



"""
Datasource-specific classes.

This is where datasource specific logic is implemented. Each mission
should have its own file with one or more classes defined.
"""
__all__ = ['CallistoSpectrogram', 'SWavesSpectrogram']

from .callisto import CallistoSpectrogram
from .swaves import SWavesSpectrogram



from datetime import datetime

import numpy as np
import pytest

from radiospectra.util import ConditionalDispatch, common_base, get_day, merge, minimal_pairs, to_signed


@pytest.fixture
def oddeven(request):
    f = ConditionalDispatch()
    # Multiply even numbers by two.
    f.add(lambda x: 2 * x, lambda x: x % 2 == 0)
    # Multiply odd numbers by three.
    f.add(lambda x: 3 * x, lambda x: x % 2 == 1)
    return f


def test_dispatch(oddeven):
    assert oddeven(2) == 4
    assert oddeven(3) == 9


def test_wrong_sig(oddeven):
    with pytest.raises(TypeError) as exc_info:
        oddeven(y=2)
    assert "There are no functions matching your input parameter signature." in str(
        exc_info.value)


def test_nocond():
    f = ConditionalDispatch()
    # Multiply even numbers by two.
    f.add(lambda x: 2 * x, lambda x: x % 2 == 0)
    with pytest.raises(TypeError) as exc_info:
        f(3)
        assert "Your input did not fulfill the condition for any function." in str(
            exc_info.value)


def test_else():
    f = ConditionalDispatch()
    # Multiply even numbers by two.
    f.add(lambda x: 2 * x, lambda x: x % 2 == 0)
    f.add(lambda x: 3 * x)
    assert f(2) == 4
    assert f(3) == 9


def test_else2():
    # This verifies else branches do not catch cases that are covered
    # by cases added later.
    f = ConditionalDispatch()
    # Because gcd(2, 3) == 1, 2 | x and 3 | x are mutually exclusive.
    f.add(lambda x: 2 * x, lambda x: x % 2 == 0)
    f.add(lambda x: 3 * x)
    f.add(lambda x: 4 * x, lambda x: x % 3 == 0)
    assert f(2) == 4
    assert f(3) == 12
    assert f(5) == 15


def test_types():
    f = ConditionalDispatch()
    f.add(lambda x: 2 * x, lambda x: x % 2 == 0, [int])
    with pytest.raises(TypeError):
        f(2.0)


def test_minimal_pairs():
    """
    This should return the pairs of elements from list1 and list2 with
    minimal difference between their values.
    """
    list1 = [0, 5, 10, 15, 20, 25]
    list2 = [3, 12, 19, 21, 26, 29]
    assert list(minimal_pairs(list1, list2)) == [(1, 0, 2), (2, 1, 2),
                                                 (4, 2, 1), (5, 4, 1)]


def test_get_day():
    end_of_day = datetime(year=2017, month=1, day=1, hour=23, minute=59, second=59,
                          microsecond=999)

    begining_of_day = get_day(end_of_day)
    assert begining_of_day.year == 2017
    assert begining_of_day.month == 1
    assert begining_of_day.day == 1
    assert begining_of_day.hour == 0
    assert begining_of_day.minute == 0
    assert begining_of_day.second == 0
    assert begining_of_day.microsecond == 0


def test_common_base():
    """
    This should return the base class common to each object in objs.
    """
    class TestA(object):
        """Base test class."""

    class TestB(TestA):
        """First inherited class."""

    class TestC(TestA):
        """Second inherited class."""
    inst_b = TestB()
    inst_c = TestC()
    objs = [inst_b, inst_c]
    assert common_base(objs) == TestA


def test_merge():
    """
    This should return a sorted (from greatest to least) merged list
    from list1 and list2.
    """
    list1 = [13, 11, 9, 7, 5, 3, 1]
    list2 = [14, 12, 10, 8, 6, 4, 2]
    result = list(merge([list1, list2]))
    assert result[::-1] == sorted(result)

    assert list(merge([[], [1], []])) == [1]


def test_to_signed():
    """
    This should return a signed type that can hold uint32 and ensure that
    an exception is raised when attempting to convert an unsigned 64 bit integer
    to an integer
    """
    assert to_signed(np.dtype('uint32')) == np.dtype('int64')

    with pytest.raises(ValueError):
        to_signed(np.dtype('uint64')) == np.dtype('int64')



from pathlib import Path
from unittest import mock

import numpy as np
import pytest

import astropy.units as u
from astropy.time import Time
from sunpy.net import Fido
from sunpy.net import attrs as a

from radiospectra.net.sources.psp import RFSClient


@pytest.fixture
def client():
    return RFSClient()


@pytest.mark.parametrize("req_wave,receivers", [
    # Completely contain the both receiver ranges
    (a.Wavelength(1*u.kHz, 25000*u.kHz), ['rfs_lfr', 'rfs_hfr']),
    # Min in lower freq and max in high freq receiver
    (a.Wavelength(20*u.kHz, 15*u.MHz), ['rfs_lfr', 'rfs_hfr']),
    # Min below and max in low freq receiver
    (a.Wavelength(1*u.kHz, 100*u.kHz), ['rfs_lfr']),
    # Min and max in low freq receiver
    (a.Wavelength(20*u.kHz, 100*u.kHz), ['rfs_lfr']),
    # Min and max in high freq receiver
    (a.Wavelength(1800*u.kHz, 18000*u.kHz), ['rfs_hfr']),
    # Min in high freq receiver and max above
    (a.Wavelength(1800*u.kHz, 20000*u.kHz), ['rfs_hfr']),
    # Min and max in the over lap
    (a.Wavelength(1.4*u.MHz, 1.5*u.MHz), ['rfs_lfr', 'rfs_hfr'])
])
def test_check_wavelength(req_wave, receivers, client):
    res = client._check_wavelengths(req_wave)
    assert set(res) == set(receivers)


@pytest.mark.remote_data
def test_fido():
    atr = a.Time('2019/10/01', '2019/10/02')
    res = Fido.search(atr, a.Instrument('rfs'))
    res0 = res[0]
    isinstance(res0.client, RFSClient)
    assert len(res0) == 4
    assert res['rfs']['Start Time'].min() == Time('2019-10-01T00:00').datetime
    assert res['rfs']['End Time'].max() == Time('2019-10-02T23:59:59.999').datetime


@pytest.fixture
def http_responces():
    paths = [Path(__file__).parent / 'data' / n for n in ['resp1.html', 'resp2.html']]
    response_htmls = []
    for p in paths:
        with p.open('r') as f:
            response_htmls.append(f.read())
    return response_htmls


@mock.patch('sunpy.util.scraper.urlopen')
def test_search_with_wavelength(mock_urlopen, client, http_responces):
    mock_urlopen.return_value.read = mock.MagicMock()
    mock_urlopen.return_value.read.side_effect = http_responces
    mock_urlopen.close = mock.MagicMock(return_value=None)
    tr = a.Time('2019/10/13', '2019/10/15')
    wr1 = a.Wavelength(1*u.kHz, 1.1*u.MHz)
    res1 = client.search(tr, wr1)

    mock_urlopen.assert_called_with(
        'https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_lfr/2019/')
    assert np.array_equal(res1[0]['Wavelength'], [10, 1700] * u.kHz)
    assert len(res1) == 3
    assert res1['Start Time'].min().datetime == Time('2019-10-13T00:00').datetime
    assert res1['End Time'].max().datetime == Time('2019-10-15T23:59:59.999').datetime
    wr2 = a.Wavelength(2*u.MHz, 20*u.MHz)
    res2 = client.search(tr, wr2)
    mock_urlopen.assert_called_with(
        'https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_hfr/2019/')
    assert np.array_equal(res2[0]['Wavelength'], [1300, 19200] * u.kHz)
    assert len(res2) == 3
    assert res2.time_range().start == Time('2019-10-13T00:00').datetime
    assert res2.time_range().end == Time('2019-10-15T23:59:59.999').datetime


@pytest.mark.remote_data
def test_get_url_for_time_range(client):
    url_start = 'https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_lfr/2019/' \
                'psp_fld_l2_rfs_lfr_20191001_v02.cdf'
    url_end = 'https://spdf.gsfc.nasa.gov/pub/data/psp/fields/l2/rfs_hfr/2019/' \
              'psp_fld_l2_rfs_hfr_20191015_v02.cdf'
    tr = a.Time('2019/10/01', '2019/10/15')
    res = client.search(tr)
    urls = [i['url'] for i in res]
    assert urls[0] == url_start
    assert urls[-1] == url_end


def test_can_handle_query(client):
    atr = a.Time('2019/10/01', '2019/11/01')
    res = client._can_handle_query(atr, a.Instrument('rfs'))
    assert res is True
    res = client._can_handle_query(atr)
    assert res is False


@pytest.mark.remote_data
def test_get(client):
    query = client.search(a.Time('2019/10/05', '2019/10/10'), a.Instrument('rfr'))
    download_list = client.fetch(query)
    assert len(download_list) == len(query)



from datetime import datetime
from unittest import mock

import numpy as np
import pytest

import astropy.units as u
from sunpy.net import Fido
from sunpy.net import attrs as a

from radiospectra.net.attrs import Spacecraft
from radiospectra.net.sources.stereo import SWAVESClient


@pytest.fixture
def client():
    return SWAVESClient()


@pytest.mark.remote_data
def test_fido():
    atr = a.Time('2010/10/01', '2010/10/02')
    res = Fido.search(atr, a.Instrument('swaves'))

    assert isinstance(res[1].client, SWAVESClient)
    assert len(res[1]) == 8


http_cont = """<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
 <html>
 <head>
   <title>Index of /data/stereo/summary/2010</title>
 </head>
 <body>
   <h1>Index of /data/stereo/summary/2010</h1>
<pre><a href="?C=N;O=D">Name</a><a href="?C=M;O=A">Last modified</a><a href="?C=S;O=A">Size</a><hr>
<a href="swaves_average_20100101_a.sav">
swaves_average_20100101_a.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100101_a_hfr.dat">
swaves_average_20100101_a_hfr.dat</a> 13-Jul-2010 20:51  3.5M
<a href="swaves_average_20100101_a_lfr.dat">
swaves_average_20100101_a_lfr.dat</a> 13-Jul-2010 20:51  548K
<a href="swaves_average_20100101_b.sav">
swaves_average_20100101_b.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100101_b_hfr.dat">
swaves_average_20100101_b_hfr.dat</a> 13-Jul-2010 20:51  3.5M
<a href="swaves_average_20100101_b_lfr.dat">
swaves_average_20100101_b_lfr.dat</a> 13-Jul-2010 20:51  548K
<a href="swaves_average_20100102_a.sav">
swaves_average_20100102_a.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100102_a_hfr.dat">
swaves_average_20100102_a_hfr.dat</a> 13-Jul-2010 20:51  3.5M
<a href="swaves_average_20100102_a_lfr.dat">
swaves_average_20100102_a_lfr.dat</a> 13-Jul-2010 20:51  548K
<a href="swaves_average_20100102_b.sav">
swaves_average_20100102_b.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100102_b_hfr.dat">
swaves_average_20100102_b_hfr.dat</a> 13-Jul-2010 20:51  3.5M
<a href="swaves_average_20100102_b_lfr.dat">
swaves_average_20100102_b_lfr.dat</a> 13-Jul-2010 20:51  548K
<a href="swaves_average_20100103_a.sav">
swaves_average_20100103_a.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100103_a_hfr.dat">
swaves_average_20100103_a_hfr.dat</a> 13-Jul-2010 20:51  3.5M
<a href="swaves_average_20100103_a_lfr.dat">
swaves_average_20100103_a_lfr.dat</a> 13-Jul-2010 20:51  548K
<a href="swaves_average_20100103_b.sav">
swaves_average_20100103_b.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100103_b_hfr.dat">
swaves_average_20100103_b_hfr.dat</a> 13-Jul-2010 20:51  3.5M
<a href="swaves_average_20100103_b_lfr.dat">
swaves_average_20100103_b_lfr.dat</a> 13-Jul-2010 20:51  548K
<a href="swaves_average_20100104_a.sav">
swaves_average_20100104_a.sav</a>     13-Jul-2010 20:51  2.0M
<a href="swaves_average_20100104_a_hfr.dat">
swaves_average_20100104_a_hfr.dat</a> 13-Jul-2010 20:52  3.5M
<a href="swaves_average_20100104_a_lfr.dat">
swaves_average_20100104_a_lfr.dat</a> 13-Jul-2010 20:52  548K
<a href="swaves_average_20100104_b.sav">
swaves_average_20100104_b.sav</a>     13-Jul-2010 20:52  2.0M
<a href="swaves_average_20100104_b_hfr.dat">
swaves_average_20100104_b_hfr.dat</a> 13-Jul-2010 20:52  3.5M
<a href="swaves_average_20100104_b_lfr.dat">
swaves_average_20100104_b_lfr.dat</a> 13-Jul-2010 20:52  548K
</pre>
</body>"""


@mock.patch('sunpy.util.scraper.urlopen')
def test_swaves_client(mock_urlopen, client):
    mock_urlopen.return_value.read = mock.MagicMock(return_value=http_cont)
    mock_urlopen.close = mock.MagicMock(return_value=None)
    atr = a.Time('2010/01/02', '2010/01/03')
    query = client.search(atr, a.Instrument('swaves'), Spacecraft('a'))
    mock_urlopen.assert_called_with('https://solar-radio.gsfc.nasa.gov/data/stereo/summary/2010/')
    assert len(query) == 8
    assert query[0]['Source'] == 'STEREO'
    assert query[0]['Spacecraft'] == 'a'
    assert np.array_equal(query[0]['Wavelength'], [10, 160] * u.kHz)
    assert query[0]['Start Time'].datetime == datetime(2010, 1, 2)
    assert query[0]['End Time'].datetime == datetime(2010, 1, 2, 23, 59, 59, 999000)
    assert query[7]['Source'] == 'STEREO'
    assert query[7]['Spacecraft'] == 'b'
    assert np.array_equal(query[7]['Wavelength'], [125, 16000] * u.kHz)
    assert query[7]['Start Time'].datetime == datetime(2010, 1, 3)
    assert query[7]['End Time'].datetime == datetime(2010, 1, 3, 23, 59, 59, 999000)


@pytest.mark.parametrize("query_wave, receivers", [(a.Wavelength(1*u.GHz, 2*u.GHz), []),
                                                   (a.Wavelength(1*u.Hz, 2*u.Hz), []),
                                                   (a.Wavelength(20*u.kHz, 150*u.kHz), ['lfr']),
                                                   (a.Wavelength(0.13*u.MHz, 15*u.MHz), ['hfr']),
                                                   (a.Wavelength(10*u.MHz, 200*u.MHz), ['hfr']),
                                                   (a.Wavelength(100*u.Hz, 100*u.kHz), ['lfr']),
                                                   (a.Wavelength(20*u.kHz, 15*u.MHz),
                                                    ['lfr', 'hfr']),
                                                   (a.Wavelength(5*u.kHz, 20*u.MHz),
                                                    ['lfr', 'hfr'])])
def test_check_wavelength(query_wave, receivers, client):
    assert set(client._check_wavelengths(query_wave)) == set(receivers)



"""
Offer a callable object that dispatches based on arbitrary conditions and
function signature. That means, whenever it is called, it finds the registered
methods that match the input's signature and then checks for user-defined
conditions and types.

First, we need to create a new ConditionalDispatch

>>> from radiospectra.util import ConditionalDispatch
>>> fun = ConditionalDispatch()

We then can start adding branches, in this case we add a branch for
even integers, in which case the function applied is a multiplication by
three.

>>> fun.add(lambda x: 3 * x, lambda x: x % 2 == 0, [int])

By adding the other branch (odd), the function can be used for all integers.
In the case of an odd integer, we double the input. Please note that the system
has no way of verifying the conditions are mutually exclusive. In some cases
it can even be useful to use not mutually exclusive conditions, in which case
the branch that was added the earliest is executed.

>>> fun.add(lambda x: 2 * x, lambda x: x % 2 == 1, [int])

We can verify this is working.

>>> fun(2)
6
>>> fun(3)
6

And that using a float, e.g., does not.

>>> fun(3.2)
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "/home/florian/Projects/sunpy/sunpy/util/cond_dispatch.py", line 128, in __call__
    "There are no functions matching your input parameter "
TypeError: There are no functions matching your input parameter signature.


We can then add a branch for floats, giving the condition None that means
that this branch is always executed for floats.

>>> fun.add(lambda y: 5 * y, None, [float])

Also note that the float branch takes y, while the integer branch takes x.
Thus, if the user explicitly passes fun(x=1) using a keyword argument, only
the integer branch is considered. This can be useful if the user wants
control over which kind of data they are passing the the function.

>>> fun(2.0)
10.0
>>> fun(y=2)
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "/home/florian/Projects/sunpy/sunpy/util/cond_dispatch.py", line 128, in __call__
    "There are no functions matching your input parameter "
TypeError: There are no functions matching your input parameter signature.
>>> fun(y=2.5)
12.5
>>> fun(x=2)
6
>>> fun(x=2.5)
Traceback (most recent call last):
  File "<stdin>", line 1, in <module>
  File "/home/florian/Projects/sunpy/sunpy/util/cond_dispatch.py", line 128, in __call__
    "There are no functions matching your input parameter "
TypeError: There are no functions matching your input parameter signature.
"""


import os
import glob
import inspect
from datetime import datetime
from itertools import chain, repeat

import numpy as np

from sunpy.util.config import get_and_create_download_dir
from sunpy.util.net import download_file

__all__ = ['run_cls', 'matches_types', 'arginize', 'correct_argspec',
           'matches_signature', 'ConditionalDispatch', 'fmt_argspec_types',
           'minimal_pairs', 'get_day', 'to_signed', 'common_base', 'merge', 'Parent']


def merge(items, key=(lambda x: x)):
    """
    Given sorted lists of iterables, return new iterable that returns
    elements of all iterables sorted with respect to key.
    """
    state = {}
    for item in map(iter, items):
        try:
            first = next(item)
        except StopIteration:
            continue
        else:
            state[item] = (first, key(first))

    while state:
        for item, (value, tk) in list(state.items()):
            # Value is biggest.
            if all(tk >= k for it, (v, k) in list(state.items())
                   if it is not item):
                yield value
                break
        try:
            n = next(item)
            state[item] = (n, key(n))
        except StopIteration:
            del state[item]


def common_base(objs):
    """
    Find class that every item of objs is an instance of.
    """
    for cls in objs[0].__class__.__mro__:
        if all(isinstance(obj, cls) for obj in objs):
            break
    return cls


def to_signed(dtype):
    """
    Return dtype that can hold data of passed dtype but is signed.
    Raise ValueError if no such dtype exists.

    Parameters
    ----------
    dtype : `numpy.dtype`
        dtype whose values the new dtype needs to be able to represent.

    Returns
    -------
    `numpy.dtype`
    """
    if dtype.kind == "u":
        if dtype.itemsize == 8:
            raise ValueError("Cannot losslessly convert uint64 to int.")
        dtype = "int{:d}".format(min(dtype.itemsize * 2 * 8, 64))
    return np.dtype(dtype)


def get_day(dt):
    """ Return datetime for the beginning of the day of given datetime. """
    return datetime(dt.year, dt.month, dt.day)


def minimal_pairs(one, other):
    """ Find pairs of values in one and other with minimal distance.
    Assumes one and other are sorted in the same sort sequence.

    Parameters
    ----------
    one, other : sequence
        Sequence of scalars to find pairs from.

    Returns
    -------
    `tuple`
         Pairs of values in `one` and `other` with minimal distance
    """
    lbestdiff = bestdiff = bestj = besti = None
    for i, freq in enumerate(one):
        lbestj = bestj

        bestdiff, bestj = None, None
        for j, o_freq in enumerate(other[lbestj:]):
            j = lbestj + j if lbestj else j
            diff = abs(freq - o_freq)
            if bestj is not None and diff > bestdiff:
                break

            if bestj is None or bestdiff > diff:
                bestj = j
                bestdiff = diff

        if lbestj is not None and lbestj != bestj:
            yield (besti, lbestj, lbestdiff)
            besti = i
            lbestdiff = bestdiff
        elif lbestdiff is None or bestdiff < lbestdiff:
            besti = i
            lbestdiff = bestdiff

    yield (besti, bestj, lbestdiff)


def run_cls(name):
    """
    run_cls("foo")(cls, \\*args, \\**kwargs) -> cls.foo(\\*args, \\**kwargs)
    """
    fun = lambda cls, *args, **kwargs: getattr(cls, name)(*args, **kwargs)  # NOQA
    fun.__name__ = str(name)
    fun.run_cls = True
    return fun


def matches_types(fun, types, args, kwargs):
    """
    See if args and kwargs match are instances of types.

    types are given in the order they are defined in the function.
    kwargs are automatically converted into that order.
    """
    return all(
        isinstance(obj, cls) for obj, cls in list(zip(
            arginize(fun, args, kwargs), types
        ))
    )


def arginize(fun, a, kw):
    """
    Turn args and kwargs into args by considering the function signature.
    """
    args, varargs, keywords, defaults = correct_argspec(fun)

    if varargs is not None:
        raise ValueError
    names = args[len(a):]
    if defaults:
        defs = dict(list(zip(args[-len(defaults):], defaults)))
    else:
        defs = {}
    return list(a) + [kw.get(name, defs.get(name, None)) for name in names]


def correct_argspec(fun):
    """
    Remove first argument if method is bound.
    """
    fullargspec = inspect.getfullargspec(fun)
    args, varargs, keywords, defaults = fullargspec[0], fullargspec[1], fullargspec[2], fullargspec[3]
    if inspect.ismethod(fun):
        args = args[1:]
    return args, varargs, keywords, defaults


def matches_signature(fun, a, kw):
    """
    Check whether function can be called with a as args and kw as kwargs.
    """
    args, varargs, keywords, defaults = correct_argspec(fun)

    if varargs is None and len(a) > len(args):
        return False
    skw = set(kw)
    sargs = set(args[len(a):])

    # There mayn't be unexpected parameters unless there is a **kwargs
    # in fun's signature.
    if keywords is None and skw - sargs != set():
        return False
    rest = set(args[len(a):]) - set(kw)

    # If there are any arguments that weren't passed but do not have
    # defaults, the signature does not match.
    defs = set() if defaults is None else set(defaults)
    if keywords is None and rest > defs:
        return False
    return True


class ConditionalDispatch(object):
    def __init__(self):
        self.funcs = []
        self.nones = []

    @classmethod
    def from_existing(cls, cond_dispatch):
        new = cls()
        new.funcs = cond_dispatch.funcs[:]
        new.nones = cond_dispatch.nones[:]
        return new

    def add_dec(self, condition):
        def _dec(fun):
            self.add(fun, condition)
            return fun
        return _dec

    def add(self, fun, condition=None, types=None, check=True):
        """
        Add fun to ConditionalDispatch under the condition that the arguments
        must match.

        If condition is left out, the function is executed for every
        input that matches the signature. Functions are considered in
        the order they are added, but ones with condition=None are
        considered as the last: that means, a function with condition
        None serves as an else branch for that signature. conditions
        must be mutually exclusive because otherwise which will be
        executed depends on the order they are added in. Function
        signatures of fun and condition must match (if fun is bound, the
        bound parameter needs to be left out in condition).
        """
        if condition is None:
            self.nones.append((fun, types))
        elif check and correct_argspec(fun) != correct_argspec(condition):
            raise ValueError(
                "Signature of condition must match signature of fun."
            )
        else:
            self.funcs.append((fun, condition, types))

    def __call__(self, *args, **kwargs):
        matched = False
        for fun, condition, types in self.funcs:
            if (matches_signature(condition, args, kwargs) and
                    (types is None or matches_types(condition, types, args, kwargs))):
                matched = True
                if condition(*args, **kwargs):
                    return fun(*args, **kwargs)
        for fun, types in self.nones:
            if (matches_signature(fun, args, kwargs) and
                    (types is None or matches_types(fun, types, args, kwargs))):
                return fun(*args, **kwargs)

        if matched:
            raise TypeError(
                "Your input did not fulfill the condition for any function."
            )
        else:
            raise TypeError(
                "There are no functions matching your input parameter "
                "signature."
            )

    def wrapper(self):
        return lambda *args, **kwargs: self(*args, **kwargs)

    def get_signatures(self, prefix="", start=0):
        """
        Return an iterator containing all possible function signatures. If
        prefix is given, use it as function name in signatures, else leave it
        out. If start is given, leave out first n elements.

        If start is -1, leave out first element if the function was
        created by run_cls.
        """
        for fun, condition, types in self.funcs:
            if start == -1:
                st = getattr(fun, 'run_cls', 0)
            else:
                st = start

            if types is not None:
                yield prefix + fmt_argspec_types(condition, types, st)
            else:
                args, varargs, keywords, defaults = correct_argspec(condition)
                args = args[st:]
                sig_str = str(inspect.signature(fun))
                if sig_str == '(cls, *args, **kwargs)':
                    sig_str = '(cls, \\*args, \\**kwargs)'
                yield prefix + sig_str

        for fun, types in self.nones:
            if types is not None:
                yield prefix + fmt_argspec_types(fun, types, st)
            else:
                args, varargs, keywords = correct_argspec(condition)
                args = args[st:]
                if sig_str == '(cls, *args, **kwargs)':
                    sig_str = '(cls, \\*args, \\**kwargs)'
                yield prefix + sig_str

    def generate_docs(self):
        fns = (item[0] for item in chain(self.funcs, self.nones))
        return '\n\n    '.join("{} -> :meth:`{}`".format(sig, fun.__name__)
                               for sig, fun in
                               # The 1 prevents the cls from incorrectly being shown in the
                               # documentation.
                               list(zip(self.get_signatures("create", -1), fns))
                               )


def fmt_argspec_types(fun, types, start=0):
    args, varargs, keywords, defaults = correct_argspec(fun)
    args = args[start:]
    types = types[start:]

    NULL = object()
    if defaults is None:
        defaults = []
    defs = chain(repeat(NULL, len(args) - len(defaults)), defaults)

    spec = []
    for key, value, type_ in zip(args, defs, types):
        # This is a work around for a bug introduced during Python 3 porting.
        # for some reason the type was being passed in as a length 1 tuple.
        # This extracts the type under that condition. SM 6/10/15
        if isinstance(type_, tuple) and len(type_) == 1:
            type_ = type_[0]
        if value is NULL:
            spec.append("{}: {}".format(key, type_.__name__))
        else:
            spec.append("{}: {} = {}".format(key, type_.__name__, value))
    if varargs is not None:
        spec.append('*{!s}'.format(varargs))
    if keywords is not None:
        spec.append('**{!s}'.format(keywords))
    return '(' + ', '.join(spec) + ')'


class Parent(object):
    _create = ConditionalDispatch()

    @classmethod
    def read(cls, filename):
        raise NotImplementedError

    @classmethod
    def read_many(cls, filenames):
        return list(map(cls.read, filenames))

    @classmethod
    def from_glob(cls, pattern):
        """ Read out files using glob (e.g., ~/BIR_2011*) pattern. Returns
        list of objects made from all matched files.
        """
        return cls.read_many(glob.glob(pattern))

    @classmethod
    def from_single_glob(cls, singlepattern):
        """ Read out a single file using glob (e.g., ~/BIR_2011*) pattern.
        If more than one file matches the pattern, raise ValueError.
        """
        matches = glob.glob(os.path.expanduser(singlepattern))
        if len(matches) != 1:
            raise ValueError("Invalid number of matches: {:d}".format(len(matches)))
        return cls.read(matches[0])

    @classmethod
    def from_files(cls, filenames):
        """ Return list of object read from given list of
        filenames. """
        filenames = list(map(os.path.expanduser, filenames))
        return cls.read_many(filenames)

    @classmethod
    def from_file(cls, filename):
        """ Return object from file. """
        filename = os.path.expanduser(filename)
        return cls.read(filename)

    @classmethod
    def from_dir(cls, directory):
        """ Return list that contains all files in the directory read in. """
        directory = os.path.expanduser(directory)
        return cls.read_many(
            os.path.join(directory, elem) for elem in os.listdir(directory)
        )

    @classmethod
    def from_url(cls, url):
        """ Return object read from URL.

        Parameters
        ----------
        url : str
            URL to retrieve the data from
        """
        path = download_file(url, get_and_create_download_dir())
        return cls.read(path)


Parent._create.add(
    run_cls('from_file'),
    lambda cls, filename: os.path.isfile(os.path.expanduser(filename)),
    [type, str], check=False
)
Parent._create.add(
    # pylint: disable=W0108
    # The lambda is necessary because introspection is performed on the
    # argspec of the function.
    run_cls('from_dir'),
    lambda cls, directory: os.path.isdir(os.path.expanduser(directory)),
    [type, str], check=False
)
# If it is not a kwarg and only one matches, do not return a list.
Parent._create.add(
    run_cls('from_single_glob'),
    lambda cls, singlepattern: ('*' in singlepattern and
                                len(glob.glob(
                                    os.path.expanduser(singlepattern))) == 1),
    [type, str], check=False
)
# This case only gets executed under the condition that the previous one wasn't.
# This is either because more than one file matched, or because the user
# explicitly used pattern=, in both cases we want a list.
Parent._create.add(
    run_cls('from_glob'),
    lambda cls, pattern: '*' in pattern and glob.glob(
        os.path.expanduser(pattern)
    ),
    [type, str], check=False
)
Parent._create.add(
    run_cls('from_files'),
    lambda cls, filenames: True,
    types=[type, list], check=False
)
Parent._create.add(
    run_cls('from_url'),
    lambda cls, url: True,
    types=[type, str], check=False
)



"""
radiospectra
============

An open-source Python library for radio spectra in solar physics.

* Homepage: https://sunpy.org
* Documentation: https://docs.radiospectra.org/en/stable/
* Source code: https://github.com/sunpy/radiospectra
"""
import sys

from .version import version as __version__

# Enforce Python version check during package import.
__minimum_python_version__ = "3.7"


class UnsupportedPythonError(Exception):
    """Running on an unsupported version of Python."""


if sys.version_info < tuple(int(val) for val in __minimum_python_version__.split('.')):
    # This has to be .format to keep backwards compatibly.
    raise UnsupportedPythonError(
        "sunpy does not support Python < {}".format(__minimum_python_version__))

__all__ = ["__version__"]
